% boolean functions are not used as "arguments"
% goals are just facts...

:- dynamic refers_unk/1.	% Keep track of references to unknown
:- dynamic current_rule_type/1.

translate(FileName):-
	assign_rule_types,
	tell(FileName),
	( (decl(rule,_,[_,_,dynamic]),!
	  ;decl(rule,_,[_,_,transition])
	  ),!, 
	
	% if we have dynamic rules
	  debugln(['%%% Situations %%%']),
	  writelist(['sit(0..pathlength).']),nl
	; true
	),
	(retractall(refers_unk(_)),!;true),
	(retractall(dynamic_rule),!;true),
	repeat,
	  ( debugln(['%%% Sorts %%%']),
	    decl(sort,S,_E),
	    t_sort(S,Rs),
	    member(R,Rs),
	    write_rule(R),nl,fail
	   
	  ; decl(var,X,Range),
	    t_set(Range,[],Range2,Rs),
	    retractall(decl(var,X,Range)),
	    asserta(decl(var,X,Range2)),
	    member(R,Rs),
	    write_rule(R),nl,fail
	    
	  ; debugln(['\n%%% Axioms %%%']),
	    decl(attr,F/N,D),
	    t_function(F/N,D),fail

	  ; debugln(['\n%%% Translation of rules %%%']),
	    decl(rule,R/Line,[H,B,Type]),
	    Type \= goal,
	    debugln(['\n% rule ',R,' (line ',Line,')']),
	    t_rule(H,B,RRs,Type),
	    member(RR,RRs),
	    write_rule(RR),nl,fail

	  ; translate_goals	  
	  ; non_concurrent_actions
	  ; action_exec
	  ; translate_unk	    
	  ; true),!,
 	told.

translate_goals :-
	decl(rule,_,[_,_,goal]),!,
	debugln(['\n%%% Translation of goals %%%']),
	decl(rule,R/Line,[H,B,goal]),
	debugln(['\n% rule ',R,' (line ',Line,')']),
	t_goal(H,B,RRs),
	member(RR,RRs),
	write_rule(RR),nl,fail.

getnumber(I,N,_):-I>N,!,fail.
getnumber(I,_N,I).
getnumber(I,N,J):- I1 is I+1,getnumber(I1,N,J).
		
performed_action(F/N,Xs,A):-
	decl(attr,F/N,[action,_,_,DV]),
	( DV=[V],!,
	  build_atom([F,'V_NOW',Xs,V],A0),
	  A = (- A0)
	
	; % no default value
	  build_atom([known,'_',F],['V_NOW',Xs],A),
	  check_refer_unk(F/N)
	).

non_concurrent_actions :- opt(conc),!,fail.		
non_concurrent_actions :-
        debugln(['\n%%% Non concurrent actions %%%']),
	non_concurrent_actions2,
	non_concurrent_actions3,
	fail.

% Avoid 2 simultaneous executions for the same action function (with different
% arguments)
non_concurrent_actions2:-
	decl(attr,F/N,[action|_]),N>0,
	var_tuple(N,'X',F/N,Xs,RangeXs),
	performed_action(F/N,Xs,AtomF1),
	var_tuple(N,'Y',F/N,Ys,RangeYs),
	performed_action(F/N,Ys,AtomF2),
	getnumber(1,N,I),
	concat_atom(['X',I],XI),
	concat_atom(['Y',I],YI),
	rule((:- [AtomF1,AtomF2,RangeXs,RangeYs,sit('V_NOW'),XI =\= YI]) ,R),
	write_rule(R),nl,
	fail.
non_concurrent_actions2.	

non_concurrent_actions3:-
	findall(F/N,decl(attr,F/N,[action|_]),Fs),
	get_pair(Fs,F/N,F2/N2),
	var_tuple(N,'X',F/N,Xs,RangeXs),
	performed_action(F/N,Xs,AtomF1),
	var_tuple(N,'Y',F2/N2,Ys,RangeYs),
	performed_action(F2/N2,Ys,AtomF2),
	rule((:- [AtomF1,AtomF2,RangeXs,RangeYs,sit('V_NOW')]) ,R),
	write_rule(R),nl,
	fail.
non_concurrent_actions3.

get_pair(Xs,X,Y):-append(_,[X|L],Xs),append(_,[Y|_],L).

translate_unk:-	
	refers_unk(_),!,
        debugln(['\n%%% Axioms for "known" %%%']),
	refers_unk(F/N),
	axiom_unkown(F/N),fail.

assign_rule_types:-
	  ( decl(rule,R,[Head,Body,dynamic]),
	    subterms([r/3],[Head,Body],RRefs),
	    get_rule_type(RRefs,Type),
	    mark(R,Type),fail
	  ;true),!.


get_rule_type(Refs,static):- checklist(static_ref,Refs),!.
get_rule_type(Refs,transition):- member(r(1,_,_),Refs),!.
get_rule_type(_Refs,dynamic).

static_ref(r(0,X,As)):-	length(As,N),decl(attr,X/N,[static|_]).
mark(R,Type):-
	retract(decl(rule,R,[H,B,_])),
	assertz(decl(rule,R,[H,B,Type])).

%%%%%%%%%%%%%%%%%%%%%%%
action_exec :- 
	\+ decl(attr,_,[action|_]),!,fail.

action_exec :- 
	debugln(['\n%%% Generate action executions %%%']),
	decl(attr,F/N,[action|_]),
	var_tuple(N,'X',F/N,Xs,RangeXs),
	decl(func,F/N,[_,_,Range,DefVal]),
	RangeV =.. [Range,'V'],
	build_atom([F,'V_NOW',Xs,'V'],FATOM),
	rule( 
	 (FATOM :- [RangeXs,RangeV,sit('V_NOW'),'V_NOW' =\= pathlength,
	            'not -' FATOM]),            Rule1
	),
	write_rule(Rule1),nl,
	rule( 
	  (-FATOM :- [RangeXs,RangeV,sit('V_NOW'),'V_NOW' =\= pathlength,
	             not FATOM]),
	   Rule2)
	  ,
	write_rule(Rule2),nl,
	at_least_a_value(F/N,DefVal),
	fail.
	
at_least_a_value(_F,[]):-!.
at_least_a_value(F/N,[_DV]):-
	check_refer_unk(F/N),
	var_tuple(N,'X',F/N,Xs,RangeXs),
	build_atom([known,'_',F],['V_NOW',Xs],KNOWN),	
	debugln(['\n% An action with default value cannot be unknown']),
	rule(( :- [not KNOWN, RangeXs, sit('V_NOW'), 'V_NOW' =\= pathlength]),Rule),
	write_rule(Rule),nl.
	
%%%%%%%%%%%%%%%%%%%%%%%

change_var(v(0,V),V):-atom(V),!.
change_var(v(X,I),V):-
	concat_atom(['V',X,'_',I],V).
change_vars(T,T1):-map_subterms(change_var,[v/2],T,T1).

change_op(A =\= B, A '!=' B):-!.
change_op(A = B, A == B):-opt(backend(lparse)),!.
change_op(A = B, A = B):-!.
change_ops(T,T1):- map_subterms(change_op,[=\= / 2, = /2],T,T1).

write_rules(Rs):-
	member(R,Rs),write_rule(R),nl,fail.
write_rules(_Rs).	

write_rule(R):-
	change_vars(R,R0),
	change_ops(R0,R1),
	(R1 = (H :- B),!,
	  write(H),write_body(B)
	; R1 = ( :- B),!, write_body(B)
	),
	write('.').


write_body([]):-!.
write_body(B):- is_list(B),!,write(':- '),writelist((,),B).
write_body(B):- write(' :- '),write(B).
	
:- dynamic varnum/1,prednum/1.
varnum(0).
prednum(0).
	
fresh_var(v(0,V)):-
	varnum(V),incr(varnum,1).
	
var_tuple(0,_I,_F,[],[]):-!.
var_tuple(N,Initial,F,Xs,[Lit|LitsYs]):-
	M is N-1,
	var_tuple(M,Initial,F,Ys,LitsYs),
	concat_atom([Initial,N],X),
	append(Ys,[X],Xs),
	decl(func,F,[_,Dom,_,_]),
	nth1(N,Dom,A),
	Lit =.. [A,X].

fresh_predicate(P):-
	prednum(X),incr(prednum,1),concat_atom([aux,X],P).
	
%%% FUNCTION DEFINITIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


range(F/N,R):-decl(func,F/N,[_,_,R,_]).

axiom_unkown(F/N):-
	decl(func,F/N,[_Type,_Ds,R,_DefVal]),
	(decl(attr,F/N,[static|_]),!,SIT=0,TLits=[]
	; SIT='V_NOW',TLits=[sit('V_NOW')]
	),
	% Axiom for "known"
	var_tuple(N,'X',F/N,Xs,RangeXs),
	build_atom([known,'_',F],[SIT,Xs],HeadK),
	build_atom([F,SIT,Xs,'V'],HATOMK),
	RangeV =.. [R,'V'],
	merge_literals([HATOMK, RangeV, TLits,RangeXs],BodyK),
	write_rule(HeadK :- BodyK),nl.

t_function(F/N,[Type,Domain,Range,DefVal]) :-
	set_current_rule_type(dynamic),
	t_sets(Domain,[],Ds,Rs0),
	t_set(Range,[],R,Rs1),	
	append(Rs0,Rs1,Rs2),
	(Rs2=[],!; debugln(['% Function definition']),write_rules(Rs2)),
	asserta(decl(func,F/N,[Type,Ds,R,DefVal])),

	var_tuple(N,'X',F/N,Xs,RangeXs),
	RangeV =.. [R,'V'],

	get_situation(F/N,0,SIT),
	(SIT='V_NOW',!,TLits=[sit('V_NOW')]; TLits=[]),
	% Unique value
	build_atom([F,SIT,Xs,'W'],HeadU),
	build_atom([F,SIT,Xs,'V'],HAtomU),
	RangeW =.. [R,'W'],
	merge_literals([HAtomU, RangeV, RangeW, 
	                'V' '!=' 'W',RangeXs,TLits],BodyU),
	
	debugln(['% Unique value']),
	write_rule(- HeadU :- BodyU),nl,
	(Type=fluent,!, SIT2=0; SIT2=SIT),
	% Default value
	( DefVal=[D],!,
	  t_scalar(D,D1,LitsD),
	  build_atom([F,SIT2,Xs,D1],HAtomD),
	  merge_literals(['not -' HAtomD,LitsD,RangeXs,TLits],BodyD),
	  debugln(['% Default value']),
	  write_rule(HAtomD :- BodyD),nl
	; true
	),
	
	% Inertia
	( Type=fluent,!,
	  build_atom([F,'V_NEXT',Xs,'V'],HAtomI),
	  transition_lits(transition,TrLits),
	  build_atom([F,'V_NOW',Xs,'V'],PrevAtomI),
	  merge_literals(['not -' HAtomI,PrevAtomI,RangeV,
	                     RangeXs,TrLits],BodyI),
	  debugln(['% Inertia']),
	  write_rule(HAtomI :- BodyI),nl	  
	; true).		

%%% SCALAR TERMS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% translation of one scalar term

t_scalar(v(0,X),v(0,X),[RangeX]):-
	!, decl(var,v(0,X),R), RangeX =.. [R,v(0,X)].
t_scalar(v(I,X),v(I,X),[]):- !.
	
t_scalar(T,T,[]) :- (atom(T),!; number(T)),!; T=v(_,_),!.


t_scalar(r(I,F,Args),V,Lits):-
	get_situation(r(I,F,Args),SIT),
	fresh_var(V), length(Args,N), range(F/N,R),
	RangeV =.. [R,V],
	t_scalar_list(Args,Args1,LitsArgs),	
	build_atom([F,SIT,Args1,V],HATOM),
	sort([HATOM,RangeV],Lits1),
	merge_set(Lits1,LitsArgs,Lits).

t_scalar(T,Tb,Lits):-
	T =.. [F,T1,T2], member(F,[+,-,*,//,mod]),!,
	t_scalar(T1,T1b,LitsT1),
	t_scalar(T2,T2b,LitsT2),
	merge_set(LitsT1,LitsT2,Lits),
	Tb =.. [F,T1b,T2b].

t_scalar(-T,-Tb,Lits):-!,t_scalar(T,Tb,Lits).
t_scalar(abs(T),abs(Tb),Lits):-!,t_scalar(T,Tb,Lits).

% translation of lists of scalar terms	
t_scalar_list([],[],[]):-!.
t_scalar_list([T|Ts],[T1|Ts1],Lits):-
	!,t_scalar(T,T1,LitsT),t_scalar_list(Ts,Ts1,LitsTs),
	merge_set(LitsT,LitsTs,Lits).

%%% LITERALS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
t_literal(T,Tb,Lits):-
	T =.. [F,T1,T2], member(F,[=,=\=,>,<,>=,<=]),!,
	t_scalar(T1,T1b,LitsT1),
	t_scalar(T2,T2b,LitsT2),
	merge_set(LitsT1,LitsT2,Lits),
	Tb =.. [F,T1b,T2b].

t_literal(not r(I,F,Ts),T1,Lits):-
	!,
	get_situation(r(I,F,Ts),SIT),
	t_scalar_list(Ts,Ts1,Lits),
	build_atom([F,SIT,Ts1,false],T1).

t_literal(r(I,F,Ts),T1,Lits):-
	!,
	get_situation(r(I,F,Ts),SIT),
	t_scalar_list(Ts,Ts1,Lits),
	build_atom([F,SIT,Ts1,true],T1).

t_literal(unknown(r(I,F,Ts)),(not Tb),Lits):-
	get_situation(r(I,F,Ts),SIT),
	t_scalar_list(Ts,Ts1,Lits),
	concat_atom([known,'_',F],Known),
	append([Known,SIT],Ts1,Atom),
	length(Ts,N),check_refer_unk(F/N),
	Tb =.. Atom.
	
t_literal(not unknown(r(I,F,Ts)),Tb,Lits):-
	get_situation(r(I,F,Ts),SIT),
	t_scalar_list(Ts,Ts1,Lits),
	concat_atom([known,'_',F],Known),
	append([Known,SIT],Ts1,Atom),
	length(Ts,N),check_refer_unk(F/N),
	Tb =.. Atom.

check_refer_unk(F/N):-refers_unk(F/N),!.
check_refer_unk(F/N):-asserta(refers_unk(F/N)).
% translation of list of literals	
	
t_literals([],[]):-!.
t_literals([L|Rest],Lits):-
	t_literal(L,L1,LitsL),
	t_literals(Rest,LitsRest),
	merge_literals([L1,LitsL,LitsRest],Lits).


get_situation(r(I,F,Args),SIT):-
	!,length(Args,N), get_situation(F/N,I,SIT).
get_situation(F/N,I,SIT):-
	current_rule_type(Type),
	( decl(attr,F/N,[static|_]),!, SIT=0	
	; (Type=(dynamic),! ; Type=transition),!,
	  (I=0,!,SIT='V_NOW'
	  ;I=1,!,SIT='V_NEXT'
	  )
	; Type=initial,!,SIT=I
	; Type=goal,!,
	  SIT=pathlength
	),
true.
%%% SETS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% t_set(A,B,A1,Rs)
%   A  - the set term
%   B  - translation of body literals of rule in which A occurs (if applicable)
%   A1 - predicate name (translation of A)
%   Rs - resulting rules

t_set(sort(A),_B,A,[]):-!.		% a sort name

% interval notation: only constants!
t_set(Ts,B,A1,Rs):- is_list(Ts),!,	% an extensional set
	expand_intervals(Ts,Ts2),
	fresh_predicate(A1),
	free_vars(Ts2,Ys),
	maplist(get_set_rule(A1,Ys,B),Ts2,Rs).

t_set(interval(I,J),B,A1,Rs):-
	!,expand_interval(I,J,L),t_set(L,B,A1,Rs).

t_set(union([],D),B,A1,Rs) :- !, t_set(D,B,A1,Rs).
t_set(union(C,[]),B,A1,Rs) :- !, t_set(C,B,A1,Rs).

t_set(union(C,D),B,A1,[(Head :- BodyC), (Head :- BodyD) | Rs]):- !,	% union
	fresh_predicate(A1),
	free_vars([C,D],Ys),
	t_set(C,B,C1,RsC),
	t_set(D,B,D1,RsD),
	append(RsC,RsD,Rs),
	fresh_var(X),	
	append(Ys,[X],Vars),
	Head =.. [A1|Vars],	
	C2 =.. [C1|Vars],
	D2 =.. [D1|Vars],
	merge_set([C2],B,BodyC),
	merge_set([D2],B,BodyD).

t_set(intersect(C,D),B,A1,[(Head :- BodyCD)|Rs]):- !,	% intersection
	fresh_predicate(A1),
	free_vars([C,D],Ys),
	t_set(C,B,C1,RsC),
	t_set(D,B,D1,RsD),
	append(RsC,RsD,Rs),
	fresh_var(X),	
	append(Ys,[X],Vars),
	Head =.. [A1|Vars],	
	C2 =.. [C1|Vars],
	D2 =.. [D1|Vars],
	sort([C2,D2],Lits),
	merge_set(Lits,B,BodyCD).

t_set(diff(C,D),B,A1,[(Head :- BodyCD)|Rs]):- !,	% difference
	fresh_predicate(A1),
	free_vars([C,D],Ys),
	t_set(C,B,C1,RsC),
	t_set(D,B,D1,RsD),
	append(RsC,RsD,Rs),
	fresh_var(X),	
	append(Ys,[X],Vars),
	Head =.. [A1|Vars],	
	C2 =.. [C1|Vars],
	D2 =.. [D1|Vars],
	sort([C2,not D2],Lits),
	merge_set(Lits,B,BodyCD).

t_set(intensional(N,T,C),B,A1,[(Head :- Body)|RsC]):- !,
							% intensional set
	fresh_predicate(A1),
	free_vars(intensional(N,T,C),Ys),
	t_scalar(T,T1,LitsT),
	append(Ys,[T1],Args),
	Head =.. [A1|Args],
	t_body(C,B,C1,RsC),
	merge_set(C1,LitsT,Body0),
	merge_set(Body0,B,Body).

get_set_rule(A1,Ys,B1,T,(Head :- Body)):-
	t_scalar(T,T1,Lits),
	append(Ys,[T1],Args),
	Head =.. [A1|Args], 
	merge_set(B1,Lits,Body).


% a list of sets

t_sets([],_B,[],[]):-!.
t_sets([T|Ts],B,[A|As],Rs):-
	t_set(T,B,A,RsT),
	t_sets(Ts,B,As,Rs0),
	merge_set(RsT,Rs0,Rs).

expand_intervals([],[]):-!.
expand_intervals([set(interval(A,B))|Xs],L):-
	!,expand_interval(A,B,L0),
	expand_intervals(Xs,L1),
	merge_set(L0,L1,L).
expand_intervals([X|Xs],L):-
	!,expand_intervals(Xs,L1),
	merge_set([X],L1,L).

expand_interval(A,B,_):- \+ (number(A), number(B)),!,fail.
expand_interval(A,B,[]):- A>B,!.
expand_interval(A,A,[A]):- !.
expand_interval(A,B,[A|L]):- !, A1 is A+1,expand_interval(A1,B,L).	

%%% SORT DEFINITIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_sort(S,RulesS) :-
	decl(sort,S,B),!,
	t_set(B,[],B1,RulesB),
	( RulesB=[],!,   % depends on another sort
	  fresh_var(X),
          Head =.. [S,X],
	  build_atom([B1,X],HATOM),
	  Body =.. [HATOM],
	  RulesS=[(Head :- Body)]
	; replace_pred(B1,S,RulesB,RulesS)
	).

t_sort(S,[(Head :- Body)]) :-
	fresh_var(X),
	Head =.. [S,X],
	build_atom([S,X,true],HATOM),
	Body =.. [HATOM].

%%% BODIES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% t_body(body(Ls,VDecls),B,Lits,Rs)
%   Ls      - literals in the body
%   VDecls  - Var. declarations in the body
%   Lits    - resulting set of LP literals
%   Rs 	    - resulting rules (for sets in var. declarations)

t_body(body(Ls,VDecls),B,Lits,Rs) :-
	t_literals(Ls,Ls1),
	t_vardecls(VDecls,B,Ls2,Rs),
	merge_set(Ls1,Ls2,Lits).
	
t_vardecls([],_B,[],[]):-!.
t_vardecls([var(Vs,A)|VDecls],B,Lits,Rs):- !,
	t_set(A,B,A1,RsA),
	maplist(build_lit(A1),Vs,VsLits),
	t_vardecls(VDecls,B,Lits0,Rs0),
	merge_set(Lits0,VsLits,Lits),
	merge_set(Rs0,RsA,Rs).	

build_lit(A1,V,L):- L =.. [A1,V].

%%% RULES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_current_rule_type(T):-
	(retractall(current_rule_type(_)),!;true),
	asserta(current_rule_type(T)).
	
t_rule(false,Body,Rs,Type):-
	!,	
	set_current_rule_type(Type),
	transition_lits(Type,TLits),
	t_body(Body,[],Lits,Rs0),
	merge_literals([Lits,TLits],Body1),
	Rs=[(:- Body1)|Rs0].
	
t_rule(r(I,F,Ts)=T0,Body,Rs,Type):-
	!,
	set_current_rule_type(Type),
	transition_lits(Type,TLits),
	t_scalar_list(Ts,Ts1,LitsTs),	
	t_scalar(T0,T0b,LitsT0),
	get_situation(r(I,F,Ts),SIT),
	build_atom([F,SIT,Ts1,T0b],Head),
	t_body(Body,[],Lits,Rs0),
	merge_literals([Lits,LitsTs,LitsT0,TLits],Lits1),
	Rule1 = (Head :- Lits1),
	
	length(Ts,N),range(F/N,Range),
	RangeT0 =.. [Range,T0b],
	merge_literals([Lits1,not RangeT0,TLits],Lits2),
	Rule2 = ( :- Lits2),
	Rs=[Rule1,Rule2|Rs0].

t_rule(in(r(I,F,Ts),A),Body,Rs,Type):-
	!,
	set_current_rule_type(Type),
	transition_lits(Type,TLits),
	t_scalar_list(Ts,Ts1,LitsTs),	
	fresh_var(V), 
	length(Ts,N), range(F/N,Range), RangeV =.. [Range,V],
	
	t_body(Body,[],LitsBody,Rs0),
	merge_literals([LitsTs,LitsBody,TLits],Lits1),
	
	t_set(A,Lits1,A1,Rs1),
	
	merge_set(Rs0,Rs1,Rs2),

	free_vars(A,Ys),
	append(Ys,[V],Args2),
	AATOM =.. [A1|Args2],

	get_situation(r(I,F,Ts),SIT),
	build_atom([F,SIT,Ts1,V],FATOM),	
	
	rule( (FATOM :- [Lits1,AATOM,'not -' FATOM,RangeV,TLits]), Rule1),
	rule( (-FATOM :- [Lits1,AATOM,not FATOM,RangeV,TLits]), Rule2),

	% P is used to check that at least one value from the set was assigned
	fresh_predicate(P),
	PATOM =.. [P|Ts1],
	rule( (PATOM :- [Lits1,AATOM,FATOM,RangeV,TLits]), Rule3),

	% note that the constraint should not refer to AATOM	
	rule( ( :- [Lits1,not PATOM,RangeV,TLits]), Rule4),
	
	Rs=[Rule1, Rule2, Rule3, Rule4|Rs2].


rule((Head :- Body),(Head :- Body2)) :-
	!,merge_literals(Body,Body2).
rule(( :- Body),( :- Body2)) :-
	merge_literals(Body,Body2).

t_goal(r(_I,F,Ts)=T0,Body,Rs):-
	!,
	set_current_rule_type(goal),
	t_scalar_list(Ts,Ts1,LitsTs),	
	t_scalar(T0,T0b,LitsT0),
	build_atom([F,pathlength,Ts1,T0b],Head),
	t_body(Body,[],Lits,Rs0),
	merge_literals([not Head,Lits,LitsTs,LitsT0],Lits1),
	Rule1 = (:- Lits1),
	Rs=[Rule1|Rs0].

transition_lits(static,[]):-!.
transition_lits(initial,[]):-!.
transition_lits(dynamic,[sit('V_NOW')]):-!.
transition_lits(transition,[sit('V_NOW'),sit('V_NEXT'),'V_NEXT'='V_NOW'+1]):-
	opt(backend(lparse)),!.
transition_lits(transition,[sit('V_NOW'),sit('V_NEXT'),'#succ(V_NOW,V_NEXT)']):-
	opt(backend(dlv)),!.
	


merge_literals([],[]):-!.
merge_literals([L|Ls],Ns):-
	is_list(L),!,merge_literals(Ls,Ms),merge_set(L,Ms,Ns).
merge_literals([L|Ls],Ns):-
	merge_literals(Ls,Ms),merge_set([L],Ms,Ns).


%%% Get the free variables in any term %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
free_vars(v(X,Y),[v(X,Y)]):-!.
free_vars(intensional(_L,T,body(Cond,VarDecls)),Vs):-
	!,free_vars(T,TVs),
	free_vars_list(Cond,CondVs),
	merge_set(TVs,CondVs,Vs0),
	findall(V,(member(var(Ws,_),VarDecls),member(V,Ws)),Vs1),
	sort(Vs1,Vs2),
	subtract(Vs0,Vs2,Vs).
free_vars(r(_,_,Args),Vs):-!,free_vars_list(Args,Vs).
free_vars([],[]):-!.
free_vars(T,Vs):-
	T =.. [_F|Args],
	free_vars_list(Args,Vs).

free_vars_list([],[]):-!.
free_vars_list([T|Ts],Vs):-
	free_vars(T,TVs),free_vars_list(Ts,TsVs),
	merge_set(TVs,TsVs,Vs).


replace_pred(_,_,[],[]):-!.
replace_pred(P1,P2,[X|Xs],[Y|Ys]):-
	!,replace_pred(P1,P2,X,Y),
	replace_pred(P1,P2,Xs,Ys).
replace_pred(P1,P2,T,T2):-
	T =.. [P1|Args],!,
	replace_pred(P1,P2,Args,Args2),
	T2=.. [P2|Args2].
replace_pred(P1,P2,T,T2):-
	T =.. [F|Args],!, 
	replace_pred(P1,P2,Args,Args2), T2=.. [F|Args2].
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

build_atom([F|Args],A):-
	flatten(Args,Args2),
	A =.. [F| Args2].

build_atom(Functor,Args,A):-
	concat_atom(Functor,F),
	flatten(Args,Args2),
	A =.. [F| Args2].

